;---------------------------------------------------------------------------------------- ---- --- -- - -  -  player constants
.define ini_indexed_mask 1
.define ini_indexmask_size 32
.define ini_relative 1
.define ini_mus_loop 1

.define max_inner_call 4

.define mus_data_adr $8000

;--------------------------------------------------------------------------------------------- ---- --- -- - -  -  description
; mus_player_nes_v1.0  by  bfox & mr287cc
;----------------------------------------
; speed-optimized player for psg-packer v1.0 (NES/Famicom (Sunsoft 5b mapper)) without self-modifyed code (can be used in ROM)
;
; there're 3 entry points:  <_mus_init> - initialise the player
;                           <_mus_play> - unpack and play one frame of music
;                           <_mus_stop> - shut down the playback
;
; max number of t-states in <_mus_play> is 421 + stack_depth*22 + quantity_of_links*(40+ini_relative*6) + ini_indexed_mask*34
;
; where  [stack_depth] is a number of entries in the music stack for the current frame;
;        [quantity_of_links] per one frame during playback;
;
; (the values in square brackets differs from frame to frame and depends on the contents of the music data structure)
;
; WARNING:
;     module can be any size and links can be along its entire length (no more than 16kb per one link, but they can be nested)
;     if your music module is greater than 8kb you should add memory bank switching in <mus_links> and <mus_com_exit> routines
; _________________________________________________________________________________________________________ ____ ___ __ _ _  _

	.export	_mus_init, _mus_stop
	.export _codesize

AY_SELECT	=$c000
AY_WRITE	=$e000

.segment "ZEROPAGE"

mus_adr:		.res	2
mus_pause:		.res	1
mus_TEMP:		.res	2
mus_stack_adr:		.res	2
mus_ini_stack:		.res	max_inner_call*3

.segment "CODE"

;----------------------------------------------------------------------------------------- ----- ---- --- -- - -  -   -
; void __fastmus_call__ _mus_init(const char *module);         initialization
; ---------------------------------------------------------------------------

.proc _mus_init: near

		ldx	#>mus_data_adr
	.if (ini_indexed_mask=1)
		lda	#<(mus_data_adr+ini_indexmask_size*2)
	.else
		lda	#<mus_data_adr
	.endif
		sta	<mus_adr
		stx	<mus_adr+1
		ldx	#0
		stx	<mus_pause

		lda	#<mus_ini_stack
		sta	<mus_stack_adr
		stx	<mus_stack_adr+1

		inx			;\ not used in player
		stx	_ppsg_flags	;/
;		bne	_mus_stop
.endproc

;----------------------------------------------------------------------------------------- ----- ---- --- -- - -  -   -
; void __fastmus_call__ _mus_stop();                             stop playing
; ---------------------------------------------------------------------------

.proc 	_mus_stop: near

		ldx  #$0d
		lda  #0
@1:
		stx  AY_SELECT
		sta  AY_WRITE
		dex
		bpl  @1
		rts
.endproc

;----------------------------------------------------------------------------------------- ----- ---- --- -- - -  -   -
; void __near__	mus_play (void);                               music playback
; ---------------------------------------------------------------------------

.proc 	_mus_play: near

		ldy	mus_pause	;3
		bne	mus_skipframe	;2

;----------------------------------------------------------------------------------------- ----- ---- --- -- - -  -   -
;11hhhhhh , llllllll , nnnnnnnn - mus_call_n - play  N frames from mus_adr-00hhhhhh llllllll
;10hhhhhh , llllllll            - call_1 - play one frame from mus_adr-00hhhhhh llllllll
;01mmmmmm , mmmmmmmm ,  ......  - playback of psg2, where mmmmmm mmmmmmmm is a binary mask of AY-registers, then values
;001nnnnn                       - number of indexed mask in table (first 32 bytes of data block)
;0001pppp               - pause16 pppp+1 (1..16)
;0000hhhh , vvvvvvvv    - playback of psg1 (0000hhhh - register number + 1, vvvvvvvv - value)
;00001111               - end of playback (shutdown AY-chip or loop track)
;00000000 , nnnnnnnn    - wait for nnnnnnnn+1 frames (put nothing to registers)
;----------------------------------------------------------------------------------------- ----- ---- --- -- - -  -   -
mus_commands:
		lda	(mus_adr),y	;5+
		bpl	mus_nolink	;3	/2

;----------------------------------------------------------------------------------------- ----- ---- --- -- - -  -   -
;11hhhhhh , llllllll , nnnnnnnn	- call_n - play  N frames from mus_adr-00hhhhhh llllllll
;10hhhhhh , llllllll		- call_1 - play one frame from mus_adr-00hhhhhh llllllll
;----------------------------------------------------------------------------------------- ----- ---- --- -- - -  -   -
mus_links:
		cmp	#$c0			;2
		  and	#$3f			;2
		  sta	<mus_TEMP		;3
		  lda	#1			;2
		  ldy	#2			;2
		bcc	mus_call_1		;2 /3
mus_call_n:
		lda	(mus_adr),y		;5+
mus_call_1:
		  ldx	mus_stack_adr		;3
		  sta	0,x			;4
		  inx				;2

		tya				;2
		adc	mus_adr			;3

		  sta	0,x			;4
		  inx				;2	_38

	.if (ini_relative=1)
		tay			;2
	.endif
		lda	#0			;2
		adc	mus_adr+1		;3

		  sta	0,x			;4
		  inx				;2
		  stx	mus_stack_adr		;3
  
	.if (ini_relative=1)

		tax			;2
		tya			;2
		ldy	#1		;2
		sec			;2
		sbc	(mus_adr),y	;5+
		sta	<mus_adr	;3
		txa			;2
		sbc	mus_TEMP	;3
		sta	<mus_adr+1	;3
	.else
		dey	;#1			;2
		lda	(mus_adr),y		;5+
;		clc
		adc	#<mus_data_adr		;2
		sta	<mus_adr		;3
		lda	<mus_TEMP		;3
		adc	#>mus_data_adr		;2
		sta	<mus_adr+1		;3
	.endif
		dey			;2
		beq	mus_commands	;3	_93r/87un (delta 6)

;----------------------------------------------------------------------------------------- ----- ---- --- -- - -  -   -
mus_skipframe:
		dec	<mus_pause	;5
		rts			;6
;----------------------------------------------------------------------------------------- ----- ---- --- -- - -  -   -
; 00000000 , nnnnnnnn - wait for nnnnnnnn+1 frames (put nothing to registers)
;----------------------------------------------------------------------------------------- ----- ---- --- -- - -  -   -
mus_pause_n:
		iny			;2
		lda	(mus_adr),y	;5+
		sta	<mus_pause	;3
		jmp	mus_com_exit	;3
;----------------------------------------------------------------------------------------- ----- ---- --- -- - -  -   -
mus_nolink:
		beq	mus_pause_n	;2 /3

		cmp	#$40		;2
		bcs	mus_psg2	;2 /3

	.if (ini_indexed_mask=1)
		cmp	#$20		;2
		bcs	mus_psg2_index	;2 /3
	.endif
		cmp	#$0F		;2
	.if (ini_mus_loop=1)
		beq	mus_loop	;2 /3
	.else
		beq	AY_stop		;....
	.endif
		bcc	mus_psg1	;2 /3

;----------------------------------------------------------------------------------------- ----- ---- --- -- - -  -   -
;0001pppp - pause pppp+1 (1..16) frames (put nothing to registers)
;----------------------------------------------------------------------------------------- ----- ---- --- -- - -  -   -
; mus_pause_16:
		and	#$0F		;2
		sta	<mus_pause	;3
		jmp	mus_com_exit	;3

	.if (ini_mus_loop=1)
;----------------------------------------------------------------------------------------- ----- ---- --- -- - -  -   -
; 00001111 - end of playback (shutdown AY-chip or loop track)
;----------------------------------------------------------------------------------------- ----- ---- --- -- - -  -   -
mus_loop:
		lda	#>(mus_data_adr)			;2
		sta	<mus_adr+1				;3
	.if (ini_indexed_mask=1)
		lda	#<(mus_data_adr+ini_indexmask_size*2)	;2
	.else
		lda	#<(mus_data_adr)			;..
	.endif
		sta	<mus_adr				;3
		bcs	mus_commands				;3
	.endif
;----------------------------------------------------------------------------------------- ----- ---- --- -- - -  -   -
; 0000hhhh , vvvvvvvv	- playback psg1 (0000hhhh - register number + 1, vvvvvvvv - value)
;----------------------------------------------------------------------------------------- ----- ---- --- -- - -  -   -
mus_psg1:
		tax			;2
		dex			;2
		jmp	mus_last_one	;3

;----------------------------------------------------------------------------------------- ----- ---- --- -- - -  -   -
; 01mmmmmm , mmmmmmmm - playback of psg2, where mmmmmm mmmmmmmm is a binary mask of AY-registers, then values
;----------------------------------------------------------------------------------------- ----- ---- --- -- - -  -   -
mus_psg2:	
		iny			;2
		sta	<mus_TEMP	;3
		lda	(mus_adr),y	;5+

	.if (ini_indexed_mask=1)

		bcs	mus_p2		;3

;----------------------------------------------------------------------------------------- ----- ---- --- -- - -  -   -
; 001nnnnn - number of indexed mask in table (first 32 bytes of data block)
;----------------------------------------------------------------------------------------- ----- ---- --- -- - -  -   -
mus_psg2_index:
		and	#$1f			;2
		asl	a			;2

		tax				;2
		lda	mus_data_adr,x		;4
		sta	<mus_TEMP		;3
		inx				;2
		lda	mus_data_adr,x		;4
	.endif
;----------------------------------------------------------------------------------------- ----- ---- --- -- - -  -   -
mus_p2:
		ldx	#0		;2
		lsr	a		;2
		sta	<mus_TEMP+1	;3

		bcc	@r1		;2 /3
		stx	AY_SELECT	;4
		iny			;2
		lda	(mus_adr),y	;5+
		sta	AY_WRITE	;4	//69
@r1:
		inx			;2
		lsr	<mus_TEMP+1	;5
		bcc	@r2		;2 /3
		stx	AY_SELECT	;4
		iny			;2
		lda	(mus_adr),y	;5+
		sta	AY_WRITE	;4	//25
@r2:	
		inx			;2
		lsr	<mus_TEMP+1	;5		
		bcc	@r3		;2 /3
		stx	AY_SELECT	;4
		iny			;2
		lda	(mus_adr),y	;5+
		sta	AY_WRITE	;4	//25
@r3:	
		inx			;2
		lsr	<mus_TEMP+1	;5		
		bcc	@r4		;2 /3
		stx	AY_SELECT	;4
		iny			;2
		lda	(mus_adr),y	;5+
		sta	AY_WRITE	;4	//25
@r4:		
		inx			;2
		lsr	<mus_TEMP+1	;5		
		bcc	@r5		;2 /3
		stx	AY_SELECT	;4
		iny			;2
		lda	(mus_adr),y	;5+
		sta	AY_WRITE	;4	//25
@r5:		
		inx			;2
		lsr	<mus_TEMP+1	;5		
		bcc	@r6		;2 /3
		stx	AY_SELECT	;4
		iny			;2
		lda	(mus_adr),y	;5+
		sta	AY_WRITE	;4	//25
@r6:	
		inx			;2
		lsr	<mus_TEMP+1	;5		
		bcc	@r7		;2 /3
		stx	AY_SELECT	;4
		iny			;2
		lda	(mus_adr),y	;5+
		sta	AY_WRITE	;4	//25
@r7:	
		inx			;2
		lsr	<mus_TEMP+1	;5		
		bcc	@r8		;2 /3
		stx	AY_SELECT	;4
		iny			;2
		lda	(mus_adr),y	;5+
		sta	AY_WRITE	;4	//25
@r8:	
		inx			;2
		lsr	<mus_TEMP	;5		
		bcc	@r9		;2 /3
		stx	AY_SELECT	;4
		iny			;2
		lda	(mus_adr),y	;5+
		sta	AY_WRITE	;4	//25
@r9:	
		inx			;2
		lsr	<mus_TEMP	;5		
		bcc	@r10		;2 /3
		stx	AY_SELECT	;4
		iny			;2
		lda	(mus_adr),y	;5+
		sta	AY_WRITE	;4	//25
@r10:	
		inx			;2
		lsr	<mus_TEMP	;5		
		bcc	@r11		;2 /3
		stx	AY_SELECT	;4
		iny			;2
		lda	(mus_adr),y	;5+
		sta	AY_WRITE	;4	//25
@r11:		
		inx			;2
		lsr	<mus_TEMP	;5		
		bcc	@r12		;2 /3
		stx	AY_SELECT	;4
		iny			;2
		lda	(mus_adr),y	;5+
		sta	AY_WRITE	;4	//25
@r12:		
		inx			;2
		lsr	<mus_TEMP	;5		
		bcc	@r13		;2 /3
		stx	AY_SELECT	;4
		iny			;2
		lda	(mus_adr),y	;5+
		sta	AY_WRITE	;4	//25
@r13:		
		inx			;2
		lsr	<mus_TEMP	;5		
		bcc	mus_com_exit	;2 /3
mus_last_one:	stx	AY_SELECT	;4
		iny			;2
		lda	(mus_adr),y	;5+
		sta	AY_WRITE	;4	//25

;----------------------------------------------------------------------------------------- ----- ---- --- -- - -  -   -
mus_com_exit:
		ldx	#<mus_ini_stack	;2	_396
mus_scanloop:
		cpx	mus_stack_adr	;3
		beq	mus_com_next	;2 /3
		dec	0,x		;6
		beq	mus_call_ret	;2 /3
		inx			;2
		inx			;2
		inx			;2
		bcc	mus_scanloop	;3	22
mus_com_next:
		tya			;2
		adc	mus_adr		;3
		sta	<mus_adr	;3
		bcc	@1		;2 /3
		inc	<mus_adr+1	;3
@1:		rts			;6	421+22*innercall

mus_call_ret:	
		lda	2,x		;4
		sta	<mus_adr+1	;3
		lda	1,x		;4
		sta	<mus_adr	;3
		stx	<mus_stack_adr	;3
		rts			;6
.endproc
;----------------------------------------------------------------------------------------- ----- ---- --- -- - -  -   -

_codesize: .word .sizeof(_mus_play)+.sizeof(_mus_init)+.sizeof(_mus_stop)
